--------------------------------------------------------------------------------
--                       Paul Scherrer Institute (PSI)
--------------------------------------------------------------------------------
-- Unit    : evr320_dpram.vhd
-- Author  : Waldemar Koprek, Section Diagnostic
--           Goran Marinkovic, Section Diagnostic
--			 Benoit Stef, Section DSP
--------------------------------------------------------------------------------
-- Copyright PSI, Section Diagnostic
--------------------------------------------------------------------------------
-- Comment : modif 02.10.2019 - numeric_std instead std_logic_unsigned
--------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

entity evr320_dpram is
   generic 
   (
      MEM_SIZE_BYTE              : integer := 2048;
      MEM_DOB_WIDTH              : integer := 32
   );
   port
   (
      -- port a
      clka                        : in    std_logic;
      ena                         : in    std_logic;
      wea                         : in    std_logic;
      addra                       : in    std_logic_vector(integer(ceil(log2(real(MEM_SIZE_BYTE))))-1 downto  0);
      dia                         : in    std_logic_vector( 7 downto  0);
      -- port b
      clkb                        : in    std_logic;
      enb                         : in    std_logic;
      addrb                       : in    std_logic_vector(integer(ceil(log2(real(MEM_SIZE_BYTE/(MEM_DOB_WIDTH/8)))))-1 downto  0);
      dob                         : out   std_logic_vector(MEM_DOB_WIDTH-1 downto  0)
   );
   
   attribute ram_style                  : string;
   attribute ram_style of evr320_dpram  : entity is "block";

end evr320_dpram;

architecture behavioral of evr320_dpram is

   type ram_type_d32 is array (MEM_SIZE_BYTE-1 downto 0) of std_logic_vector(7 downto  0);
   shared variable RAM            : ram_type_d32;
   
   type ram_type_d64 is array ((MEM_SIZE_BYTE/2)-1 downto 0) of std_logic_vector(7 downto  0);
   shared variable RAM_ODD        : ram_type_d64;
   shared variable RAM_EVEN       : ram_type_d64;

begin

   -----------------------------------------------------------------------------
   dob_32bit: if MEM_DOB_WIDTH = 32 generate
   -----------------------------------------------------------------------------
      process (clka)
      begin
         if rising_edge(clka) then
            if (ena = '1') then
               if (wea = '1') then
                  RAM(to_integer(unsigned(addra))) := dia;
               end if;
            end if;
         end if;
      end process;

      process (clkb)
      	variable concat7to0_v 	: std_logic_vector(addrb'high+2 downto 0);
      	variable concat15to8_v 	: std_logic_vector(addrb'high+2 downto 0);
      	variable concat23to16_v : std_logic_vector(addrb'high+2 downto 0);
      	variable concat31to24_v : std_logic_vector(addrb'high+2 downto 0);
      begin
         if rising_edge(clkb) then
         	if (enb = '1') then
         		concat7to0_v 	:= addrb & "00";	
         		concat15to8_v 	:= addrb & "01";	
         		concat23to16_v 	:= addrb & "10";	
         		concat31to24_v 	:= addrb & "11";
         		--	
               	dob( 7 downto  0)     <= RAM(to_integer(unsigned(  concat7to0_v)));
               	dob(15 downto  8)     <= RAM(to_integer(unsigned( concat15to8_v)));
               	dob(23 downto 16)     <= RAM(to_integer(unsigned(concat23to16_v)));
               	dob(31 downto 24)     <= RAM(to_integer(unsigned(concat31to24_v)));
            end if;
         end if;
      end process;
   end generate dob_32bit;
      
   -----------------------------------------------------------------------------
   dob_64bit: if MEM_DOB_WIDTH = 64 generate
   -----------------------------------------------------------------------------
      process (clka)
      begin
         if rising_edge(clka) then
            if (ena = '1') then
               if (wea = '1') then
                  if (addra(0) = '1') then
                     RAM_ODD (to_integer(unsigned(addra(addra'high downto 1)))) := dia;
                  else
                     RAM_EVEN(to_integer(unsigned(addra(addra'high downto 1)))) := dia;
                  end if;
               end if;
            end if;
         end if;
      end process;
      
      process (clkb)
      	variable concat7to0_v 	: std_logic_vector(addrb'high+2 downto 0);
      	variable concat15to8_v 	: std_logic_vector(addrb'high+2 downto 0);
      	variable concat23to16_v : std_logic_vector(addrb'high+2 downto 0);
      	variable concat31to24_v : std_logic_vector(addrb'high+2 downto 0);
      	variable concat39to32_v : std_logic_vector(addrb'high+2 downto 0);
      	variable concat47to40_v : std_logic_vector(addrb'high+2 downto 0);
      	variable concat55to48_v : std_logic_vector(addrb'high+2 downto 0);
      	variable concat63to56_v : std_logic_vector(addrb'high+2 downto 0);
      begin
         if rising_edge(clkb) then
         	if (enb = '1') then
         		concat7to0_v 	:=addrb & "00";
				concat15to8_v   :=addrb & "00";
				concat23to16_v  :=addrb & "01";
				concat31to24_v  :=addrb & "01";
				concat39to32_v  :=addrb & "10";
				concat47to40_v  :=addrb & "10";
				concat55to48_v  :=addrb & "11";
				concat63to56_v	:=addrb & "11";
         		--
				dob( 7 downto  0)     <= RAM_EVEN(to_integer(unsigned(concat7to0_v 	)));
				dob(15 downto  8)     <= RAM_ODD (to_integer(unsigned(concat15to8_v )));
				dob(23 downto 16)     <= RAM_EVEN(to_integer(unsigned(concat23to16_v)));
				dob(31 downto 24)     <= RAM_ODD (to_integer(unsigned(concat31to24_v)));
				dob(39 downto 32)     <= RAM_EVEN(to_integer(unsigned(concat39to32_v)));
				dob(47 downto 40)     <= RAM_ODD (to_integer(unsigned(concat47to40_v)));
				dob(55 downto 48)     <= RAM_EVEN(to_integer(unsigned(concat55to48_v)));
				dob(63 downto 56)     <= RAM_ODD (to_integer(unsigned(concat63to56_v)));
            end if;
         end if;
      end process;
   end generate dob_64bit;

end behavioral;


--------------------------------------------------------------------------------
-- End of file
--------------------------------------------------------------------------------