 ---------------------------------------------------------------------------
--                       Paul Scherrer Institute (PSI)
-- ---------------------------------------------------------------------------
-- Unit    : evr320_ifc1210_wrapper.vhd
-- Author  : Patric Bucher
-- ---------------------------------------------------------------------------
-- Copyright© PSI, Section DSV
-- ---------------------------------------------------------------------------
-- Comment : Wraps evr320 decoder together with GTX component and TMEM registers.
-- ---------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library tosca2;
use tosca2.tosca2_glb_pkg.all;

use work.evr320_pkg.all;
use work.v6vlx_gtxe1_pkg.all;


entity evr320_ifc1210_wrapper is
  generic(
    g_MGT_LOCATION                : string  				:= "GTXE1_X0Y16"; -- "GTXE1_X0Y0" to "GTXE1_X0Y11" | "GTXE1_X0Y16" to "GTXE1_X0Y19"
    g_FACILITY                    : string  				:= "SFEL";        -- "HIPA" | "SFEL"
    g_EVENT_RECORDER              : boolean 				:= false;         -- enable/disable Event Recorder functionality
    g_XUSER_CLK_FREQ              : natural					:= 125000000      -- Xuser Clk Frequency in Hz
  );
  port(      
    -- ------------------------------------------------------------------------
    -- Debug interface
    -- ------------------------------------------------------------------------
    debug_clk                     	: out   std_logic;
    debug                         	: out   std_logic_vector(127 downto   0);
    -- ------------------------------------------------------------------------
    -- TOSCA2 TMEM Interface (xuser clock domain, 100-250MHz)
    -- ------------------------------------------------------------------------
    xuser_CLK						: in    std_logic;
    xuser_RESET						: in    std_logic;
    xuser_TMEM_ENA					: in    std_logic;
    xuser_TMEM_WE					: in    std_logic_vector( 7 downto 0);
    xuser_TMEM_ADD					: in    std_logic_vector(13 downto 3);
    xuser_TMEM_DATW					: in    std_logic_vector(63 downto 0);
    xuser_TMEM_DATR					: out   std_logic_vector(63 downto 0);
    -- ------------------------------------------------------------------------
    -- MGT Interface
    -- ------------------------------------------------------------------------
    mgt_refclk_i                  	: in    std_logic;                         -- MGT Reference Clock
    mgt_sfp_los_i                 	: in    std_logic;                         -- SFP Loss of Signal (light on receiver)
    mgt_rx_n                      	: in    std_logic;                         -- MGT RX N
    mgt_rx_p                      	: in    std_logic;                         -- MGT RX P
    mgt_tx_n                     	: out   std_logic;                         -- MGT TX N
    mgt_tx_p                      	: out   std_logic;                         -- MGT TX P
    mgt_status_o                  	: out   std_logic_vector(31 downto 0);     -- MGT Status
    mgt_control_i                 	: in    std_logic_vector(31 downto 0);     -- MGT Control      
    ---------------------------------------------------------------------------
    -- User interface MGT clock
    ---------------------------------------------------------------------------
    clk_evr_o                     	: out   std_logic;                         -- Recovered parallel clock from MGT
    usr_events_o                  	: out   std_logic_vector( 3 downto  0);    -- User defined event pulses with one clock cycle length
    usr_events_ext_o              	: out   std_logic_vector( 3 downto  0);    -- User defined event pulses with four clock cycle length
    sos_event_o                   	: out   std_logic;                         -- Start-of-Sequence Event
    --------------------------------------------------------------------------
    -- Decoder axi stream interface, User clock
    --------------------------------------------------------------------------
    stream_clk_i                  	: in    std_logic := '0';
    stream_data_o                 	: out   std_logic_vector(7 downto 0);  
    stream_addr_o                 	: out   std_logic_vector(10 downto 0);
    stream_valid_o                	: out   std_logic
  );
  
end evr320_ifc1210_wrapper;



architecture rtl of evr320_ifc1210_wrapper is

  -- --------------------------------------------------------------------------
  -- Parameters
  -- --------------------------------------------------------------------------
  constant c_TOSCA2_DATA_WIDTH            : integer := 64;
 -- constant c_EVR_REG64_COUNT              : integer := 16;      -- unused, only documentation
 -- constant c_EVR_MEM_SIZE                 : integer := 16384;   -- unused, only documentation
  

  -- --------------------------------------------------------------------------
  -- Signal definitions
  -- --------------------------------------------------------------------------
  signal clk_evr                          : std_logic;
--signal clk_evr_monitor                  : std_logic; -- for debugging
  signal rst_evr                          : std_logic;
  
  signal mgt_control                      : std_logic_vector(31 downto  0) := (others => '0');
  signal mgt_status                       : std_logic_vector(31 downto  0);

  signal mgt_rx_data                      : std_logic_vector(15 downto  0);
  signal mgt_rx_charisk                   : std_logic_vector( 1 downto  0);
  signal mgt_lossofsync                   : std_logic;
  signal mgt_reset_tmem_evr               : std_logic; -- for legacy reasons, ifc1210 mgt control is in tmem_psi_generic part
  
  signal mem_clk                          : std_logic;
  signal mem_addr_evr                     : std_logic_vector(11 downto 0);  
  signal mem_addr_tosca                   : std_logic_vector(10 downto 0); 
  signal mem_data                         : std_logic_vector(c_TOSCA2_DATA_WIDTH-1 downto 0);
  
  signal evr_params                       : typ_evr320_params;
  signal evr_params_sync                  : typ_evr320_params;
  signal evr_params_xuser                 : typ_evr320_params;
  signal event_recorder_status            : typ_evt_rec_status;
  signal event_recorder_control           : typ_evt_rec_ctrl;
  signal event_recorder_control_sync      : typ_evt_rec_ctrl;
  signal event_recorder_control_xuser     : typ_evt_rec_ctrl;
  
  signal evr_frequency                    : std_logic_vector(31 downto 0) := (others => '0');
  
  signal debug_data                       : std_logic_vector(127 downto 0);


  -- --------------------------------------------------------------------------
  -- Attribute definitions
  -- --------------------------------------------------------------------------
  attribute keep                          : string;
  attribute keep of clk_evr               : signal is "TRUE"; 
  attribute keep of debug_data            : signal is "TRUE";
  
  
  
-- ----------------------------------------------------------------------------
-- ----------------------------------------------------------------------------
-- ////////////////////           Main Body           /////////////////////////
-- ----------------------------------------------------------------------------
-- ----------------------------------------------------------------------------
begin

  -- --------------------------------------------------------------------------
  -- static signal assignments
  -- --------------------------------------------------------------------------
  mgt_lossofsync                  <= mgt_status(15);
  rst_evr                         <= mgt_status(15);
  mem_addr_evr                    <= '0' & mem_addr_tosca;

  mgt_control(c_GTXRESET)         <= mgt_control_i(c_GTXRESET) or mgt_sfp_los_i or mgt_reset_tmem_evr;
  mgt_control( 4 downto 1)        <= mgt_control_i( 4 downto 1);
  mgt_control(c_RXCDRRESET)       <= mgt_control_i(c_RXCDRRESET);
  mgt_control(31 downto 6)        <= mgt_control_i(31 downto 6);


  -- --------------------------------------------------------------------------
  -- Synchronisation to EVR Clock
  -- --------------------------------------------------------------------------
  prc_sync_evr: process(clk_evr)
  begin
    if rising_edge(clk_evr) then
      ---
      evr_params_sync               <= evr_params_xuser;
      evr_params                    <= evr_params_sync;
      ---
      event_recorder_control_sync   <= event_recorder_control_xuser;
      event_recorder_control        <= event_recorder_control_sync;
      ---
    end if;
  end process;
  

  -- --------------------------------------------------------------------------
  -- EVR320 Decoder
  -- --------------------------------------------------------------------------
  evr320_decoder_inst: entity work.evr320_decoder
    generic map(
      EVENT_RECORDER              => g_EVENT_RECORDER,
      MEM_DATA_WIDTH              => c_TOSCA2_DATA_WIDTH )
    port map(
      -- Debug interface
      debug_clk                   => debug_clk,
      debug                       => debug_data,
      -- GTX parallel interface
      i_mgt_rst                   => mgt_lossofsync,
      i_mgt_rx_clk                => clk_evr,
      i_mgt_rx_data               => mgt_rx_data,
      i_mgt_rx_charisk            => mgt_rx_charisk,
      -- User interface CPU clock
      i_usr_clk                   => mem_clk,
      i_evr_params                => evr_params,
      o_event_recorder_stat       => event_recorder_status,
      i_event_recorder_ctrl       => event_recorder_control,
      i_mem_addr                  => mem_addr_evr,
      o_mem_data                  => mem_data,
      -- user stream interface, user clock
      i_stream_clk                => stream_clk_i,
      o_stream_data               => stream_data_o, 
      o_stream_addr               => stream_addr_o, 
      o_stream_valid              => stream_valid_o,
      -- User interface MGT clock
      o_usr_events                => usr_events_o,
      o_usr_events_ext            => usr_events_ext_o,
      o_sos_event                 => sos_event_o
   );
   

  -- --------------------------------------------------------------------------
  -- MGT Wrapper for GTX Virtex-6
  -- --------------------------------------------------------------------------
  mgt_wrapper_inst: entity work.v6vlx_gtxe1_wrapper
    generic map(
      g_MGT_LOCATION              => g_MGT_LOCATION,
      g_FACILITY                  => g_FACILITY )
    port map(   
      -- MGT serial interface   
      i_mgt_refclk                => mgt_refclk_i,
      o_mgt_refclk                => open,
      i_mgt_rx_p                  => mgt_rx_p,
      i_mgt_rx_n                  => mgt_rx_n,
      o_mgt_tx_p                  => mgt_tx_p,
      o_mgt_tx_n                  => mgt_tx_n,
      -- MGT parallel interface   
      o_mgt_status                => mgt_status,
      i_mgt_control               => mgt_control,
      o_mgt_recclk                => clk_evr,
      o_mgt_rx_data               => mgt_rx_data,
      o_mgt_rx_charisk            => mgt_rx_charisk
    );
  

  -- --------------------------------------------------------------------------
  -- TMEM 
  -- --------------------------------------------------------------------------  
  evr320_tmem_inst: entity work.evr320_tmem
  port map(
    -- TOSCA2 TMEM Interface 
    xuser_CLK                     => xuser_CLK,         
    xuser_RESET                   => xuser_RESET,       
    xuser_TMEM_ENA                => xuser_TMEM_ENA,    
    xuser_TMEM_WE                 => xuser_TMEM_WE,        
    xuser_TMEM_ADD                => xuser_TMEM_ADD,    
    xuser_TMEM_DATW               => xuser_TMEM_DATW,   
    xuser_TMEM_DATR               => xuser_TMEM_DATR,   
    -- EVR320 Memory/Parameter Interface
    evr_params_o                  => evr_params_xuser,
    evr_frequency_i               => evr_frequency,
    evr_evt_rec_status_i          => event_recorder_status,
    evr_evt_rec_control_o         => event_recorder_control_xuser,
    mgt_status_i                  => mgt_status,        
    mgt_reset_o                   => mgt_reset_tmem_evr,          
    mem_clk_o                     => mem_clk,
    mem_addr_o                    => mem_addr_tosca,
    mem_data_i                    => mem_data
  );
  
  -- --------------------------------------------------------------------------
  -- Measure EVR Clock (based on xuser_CLK)
  -- --------------------------------------------------------------------------
  clock_meas_inst : entity work.psi_common_clk_meas
  generic map (
    MasterFrequency_g  => g_XUSER_CLK_FREQ,
    MaxMeasFrequency_g => 150000000
  )
  port map (
    ClkMaster   => xuser_CLK,
    Rst         => xuser_RESET,
    ClkTest     => clk_evr,
    FrequencyHz => evr_frequency
  );

  -- --------------------------------------------------------------------------
  -- port mapping
  -- --------------------------------------------------------------------------
  clk_evr_o               <= clk_evr;
  mgt_status_o            <= mgt_status;
  debug                   <= debug_data;
    
end rtl;
-- ----------------------------------------------------------------------------
-- ////////////////////////////////////////////////////////////////////////////
-- ----------------------------------------------------------------------------
